# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Tests for :mod:`igwn_auth_utils.scitokens`.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import os
import time
from functools import partial
from unittest import mock

from scitokens import (
    __version__ as scitokens_version,
    SciToken,
)
from scitokens.scitokens import InvalidPathError

import pytest

from .. import scitokens as igwn_scitokens
from ..error import IgwnAuthError

ISSUER = "local"
_SCOPE_PATH = "/igwn_auth_utils"
READ_AUDIENCE = "igwn_auth_utils"
READ_SCOPE = "read:{}".format(_SCOPE_PATH)
WRITE_AUDIENCE = "igwn_auth_utils2"
WRITE_SCOPE = "write:{}".format(_SCOPE_PATH)


def _os_error(*args, **kwargs):
    raise OSError


def _create_token(
    key=None,
    iss=ISSUER,
    aud=READ_AUDIENCE,
    scope=READ_SCOPE,
    **kwargs
):
    """Create a token
    """
    if key:
        from scitokens.utils.keycache import KeyCache
        keycache = KeyCache.getinstance()
        keycache.addkeyinfo(iss, "test_key", key.public_key())
    now = int(time.time())
    token = SciToken(key=key, key_id="test_key")
    token.update_claims({
        "iat": now,
        "nbf": now,
        "exp": now + 86400,
        "iss": iss,
        "aud": aud,
        "scope": scope,
    })
    token.update_claims(kwargs)
    return token


def _write_token(token, path):
    with open(path, "wb") as file:
        file.write(token.serialize(lifetime=86400))


@pytest.fixture
def rtoken(private_key):
    return _create_token(
        key=private_key,
        scope=READ_SCOPE,
    )


@pytest.fixture
def wtoken(private_key):
    return _create_token(
        key=private_key,
        aud=WRITE_AUDIENCE,
        scope=WRITE_SCOPE,
    )


@pytest.fixture
def rwtoken(private_key):
    return _create_token(
        key=private_key,
        aud=READ_AUDIENCE,
        scope=" ".join((READ_SCOPE, WRITE_SCOPE)),
    )


@pytest.fixture
def rtoken_path(rtoken, tmp_path):
    token_path = tmp_path / "token.use"
    _write_token(rtoken, token_path)
    return token_path


@pytest.fixture
def condor_creds_path(rtoken, wtoken, tmp_path):
    for token, name in (
        (rtoken, "read.use"),
        (wtoken, "write.use"),
    ):
        _write_token(token, tmp_path / name)
    return tmp_path


# -- test utilities -----------------------------

def assert_tokens_equal(a, b):
    _assert_claims_equal(a, b)
    _assert_claims_equal(b, a)


_SKIP_ASSERT_CLAIMS = {
    "exp",
    "iat",
    "nbf",
}


def _assert_claims_equal(a, b):
    for claim, value in a.claims():
        if claim in _SKIP_ASSERT_CLAIMS:
            continue
        assert b.get(claim) == value


# -- tests --------------------------------------

@pytest.mark.parametrize(("scope", "validity"), (
    (READ_SCOPE, True),  # read scope matches token
    (WRITE_SCOPE, False),  # write scope doesn't
    (None, True),  # accept any scope
))
def test_is_valid_token(rtoken, scope, validity):
    assert igwn_scitokens.is_valid_token(
        rtoken,
        READ_AUDIENCE,
        scope,
    ) is validity


@pytest.mark.parametrize(("scope", "validity"), [
    (READ_SCOPE, True),
    (WRITE_SCOPE, True),
    ([READ_SCOPE, WRITE_SCOPE], True),
    ("other", False),  # single scope, no match
    ([READ_SCOPE, "other"], False),  # partial match
])
def test_is_valid_token_multiple_scopes(rwtoken, scope, validity):
    assert igwn_scitokens.is_valid_token(
        rwtoken,
        READ_AUDIENCE,
        scope,
    ) is validity


def test_is_valid_token_invalid_path(rtoken):
    with pytest.raises(InvalidPathError):
        igwn_scitokens.is_valid_token(rtoken, READ_AUDIENCE, "read")


def test_is_valid_token_serialized(rtoken, public_pem):
    _deserialize_local = partial(
        SciToken.deserialize,
        audience=READ_AUDIENCE,
        insecure=True,
        public_key=public_pem,
    )
    with mock.patch("scitokens.SciToken.deserialize", _deserialize_local):
        assert igwn_scitokens.is_valid_token(
            rtoken.serialize(lifetime=86400),
            READ_AUDIENCE,
            READ_SCOPE,
        )


def test_is_valid_token_serialized_false():
    assert igwn_scitokens.is_valid_token(
        "bad",
        READ_AUDIENCE,
        READ_SCOPE,
    ) is False


@pytest.mark.parametrize(("issuer", "result"), [
    (None, True),
    ("local", True),
    ("other", False),
])
def test_is_valid_token_issuer(rtoken, issuer, result):
    """Test that `igwn_auth_utils.scitokens.is_valid_token` enforces
    ``issuer`` properly.
    """
    assert igwn_scitokens.is_valid_token(
        rtoken,
        READ_AUDIENCE,
        None,
        issuer=issuer,
    ) is result


def test_is_valid_token_invalidauthorizationresources(private_key):
    """Check that `is_valid` doesn't fall over tokens with invalid scopes.

    Regression test for
    <https://git.ligo.org/computing/igwn-auth-utils/-/issues/15>.
    """
    token = _create_token(
        key=private_key,
        scope=" ".join(["test.auth:relative_path", READ_SCOPE])
    )
    assert not igwn_scitokens.is_valid_token(
        token,
        token["aud"],
        READ_SCOPE,
        timeleft=-1e9,
    )


def test_is_valid_token_warn(rtoken):
    """Check that `is_valid_token` emits warnings when asked.
    """
    with pytest.warns(
        UserWarning,
        match=f"Validator rejected value of '{rtoken['iss']}' for claim 'iss'",
    ):
        assert igwn_scitokens.is_valid_token(
            rtoken,
            READ_AUDIENCE,
            None,
            issuer="something else",
            warn=True,
        ) is False


@pytest.mark.parametrize("include_any", (False, True))
@pytest.mark.parametrize(("url", "aud"), (
    # basic
    ("https://example.com/data", ["https://example.com"]),
    # no scheme
    ("example.com", ["https://example.com"]),
    # port
    ("https://example.com:443/data/test", ["https://example.com"]),
    # HTTP
    ("http://example.com:443/data/test", ["http://example.com"]),
))
def test_target_audience(url, aud, include_any):
    if include_any:
        aud += ["ANY"]
    assert igwn_scitokens.target_audience(url, include_any=include_any) == aud


def test_load_token_file(rtoken_path, rtoken, public_pem):
    assert_tokens_equal(
        igwn_scitokens.load_token_file(
            rtoken_path,
            audience=READ_AUDIENCE,
            insecure=True,
            public_key=public_pem,
        ),
        rtoken,
    )


@mock.patch.dict("os.environ")
@pytest.mark.parametrize("envname", (
    "SCITOKEN",
    "BEARER_TOKEN",
))
def test_find_token_env_scitoken(rtoken, public_pem, envname):
    os.environ[envname] = rtoken.serialize(lifetime=86400).decode("utf-8")
    assert_tokens_equal(
        igwn_scitokens.find_token(
            audience=READ_AUDIENCE,
            scope=READ_SCOPE,
            insecure=True,
            public_key=public_pem,
        ),
        rtoken,
    )


@mock.patch.dict("os.environ")
@pytest.mark.parametrize("envname", (
    "SCITOKEN_FILE",
    "BEARER_TOKEN_FILE",
))
def test_find_token_env_scitoken_file(
    rtoken,
    wtoken,
    rtoken_path,
    public_pem,
    envname,
):
    # set the wrong token as SCITOKEN
    os.environ["SCITOKEN"] = wtoken.serialize(lifetime=86400).decode("utf-8")
    # and the correct token as SCITOKEN_FILE
    os.environ[envname] = str(rtoken_path)
    # and make sure we get the correct token back
    assert_tokens_equal(
        igwn_scitokens.find_token(
            audience=READ_AUDIENCE,
            scope=READ_SCOPE,
            insecure=True,
            public_key=public_pem,
            skip_errors=True,
        ),
        rtoken,
    )


@mock.patch.dict("os.environ")
def test_find_token_condor_creds(
    rtoken,
    wtoken,
    public_pem,
    condor_creds_path,
):
    os.environ["_CONDOR_CREDS"] = str(condor_creds_path)
    for token, aud, scope in (
        (rtoken, READ_AUDIENCE, READ_SCOPE),
        (wtoken, WRITE_AUDIENCE, WRITE_SCOPE),
    ):
        assert_tokens_equal(
            igwn_scitokens.find_token(
                audience=aud,
                scope=scope,
                insecure=True,
                public_key=public_pem,
                skip_errors=True,
            ),
            token,
        )


@pytest.mark.parametrize(("audience", "msg"), [
    (READ_AUDIENCE, "could not find a valid SciToken"),
    (WRITE_AUDIENCE, (
        "could not find a valid SciToken" if scitokens_version >= "1.7.3"
        else "(Invalid audience|Audience doesn't match)"
    )),
])
@mock.patch.dict("os.environ")
# make sure a real token doesn't get in the way
@mock.patch("igwn_auth_utils.scitokens.SciToken.discover", _os_error)
def test_find_token_error(rtoken, public_pem, audience, msg):
    # token with the wrong claims
    os.environ["SCITOKEN"] = rtoken.serialize().decode("utf-8")
    # check that we get an error
    with pytest.raises(
        IgwnAuthError,
        match=msg,
    ):
        igwn_scitokens.find_token(
            audience,
            WRITE_SCOPE,
            insecure=True,
            public_key=public_pem,
            skip_errors=False,
        )


@mock.patch.dict("os.environ")
@pytest.mark.parametrize(("skip_errors", "message"), (
    (False, "Issuer is not over HTTPS"),
    (True, "could not find a valid SciToken"),
))
def test_find_token_skip_errors(rtoken, skip_errors, message):
    """Check that the ``skip_errors`` keyword for `find_token()` works
    """
    # configure a valid token (wrong claims) **HOWEVER**
    # don't add the necessary keyword arguments that would
    # enable `deserialize` to actually work
    # (see `test_find_token_error` above for that)
    os.environ["SCITOKEN"] = rtoken.serialize().decode("utf-8")

    # check that we get the normal error
    with pytest.raises(
        IgwnAuthError,
        match=message,
    ):
        igwn_scitokens.find_token(
            READ_AUDIENCE,
            READ_SCOPE,
            skip_errors=skip_errors,
        )


@mock.patch.dict("os.environ")
def test_find_token_warn(rtoken, rtoken_path, public_pem):
    os.environ["SCITOKEN"] = "blah"
    os.environ["SCITOKEN_FILE"] = str(rtoken_path)
    with pytest.warns(
        UserWarning,
        match="InvalidTokenFormat",
    ):
        assert_tokens_equal(
            igwn_scitokens.find_token(
                audience=READ_AUDIENCE,
                scope=READ_SCOPE,
                insecure=True,
                public_key=public_pem,
                skip_errors=True,
                warn=True,
            ),
            rtoken,
        )


@mock.patch.dict("os.environ")
def test_find_condor_creds_no_env(tmp_path):
    """Check that `_find_condor_creds_token_paths()` handles missing
    ``_CONDOR_CREDS`` environment variable
    """
    assert not list(igwn_scitokens._find_condor_creds_token_paths())


@mock.patch.dict("os.environ")
def test_find_condor_creds_dir_missing(tmp_path):
    """Check that `_find_condor_creds_token_paths()` handles
    ``_CONDOR_CREDS`` pointing at a directory that doesn't exist
    """
    os.environ["_CONDOR_CREDS"] = str(tmp_path / "_condor_creds")
    assert not list(igwn_scitokens._find_condor_creds_token_paths())


@mock.patch.dict("os.environ")
def test_find_condor_creds_dir_empty(tmp_path):
    """Check that `_find_condor_creds_token_paths()` handles
    ``_CONDOR_CREDS`` pointing at an empty directory
    """
    os.environ["_CONDOR_CREDS"] = str(tmp_path)
    assert not list(igwn_scitokens._find_condor_creds_token_paths())


def test_token_authorization_header(rtoken):
    """Check that `token_authorization_header` works
    """
    expected = "Bearer {}".format(rtoken.serialize().decode("utf-8"))

    # reset _serialized_token attribute
    rtoken._serialized_token = None

    # do it once to check that the call to token.serialize() works
    a = igwn_scitokens.token_authorization_header(rtoken)
    assert a == expected

    # do it again to check that the use of _serialized_token attr works
    assert igwn_scitokens.token_authorization_header(rtoken) == a



# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Utility functions for discovering valid scitokens
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import os
import warnings
from pathlib import Path
from urllib.parse import urlparse

from jwt import (
    InvalidAudienceError,
    InvalidTokenError,
)
from scitokens import (
    Enforcer,
    SciToken,
)
from scitokens.utils.errors import SciTokensException
from scitokens.scitokens import InvalidAuthorizationResource

from .error import IgwnAuthError

TOKEN_ERROR = (
    InvalidAudienceError,
    InvalidTokenError,
    SciTokensException,
)


WINDOWS = os.name == "nt"

# -- utilities --------------


def is_valid_token(
    token,
    audience,
    scope,
    issuer=None,
    timeleft=600,
    warn=False,
):
    """Test whether a ``token`` is valid according to the given claims.

    Parameters
    ----------
    token : `scitokens.SciToken`, `str`
        The token object, or serialisation, to test

    audience : `str`, `list` or `str`
        The audience(s) to accept.

    scope : `str`, `list`
        One or more scopes to enforce.

    timeleft : `float`
        The amount of time remaining (in seconds, from the `exp` claim)
        to require.

    issuer : `str`
        The value of the `iss` claim to enforce.

    warn : `bool`
        If `True`, emit a warning when a token fails the enforcer test,
        useful in debugging bad tokens.

    Returns
    -------
    valid : `bool`
        `True` if the input ``token`` matches the required claims,
        otherwise `False`.
    """
    # if given a serialised token, deserialise it now
    if isinstance(token, (str, bytes)):
        try:
            token = SciToken.deserialize(token)
        except (InvalidTokenError, SciTokensException):
            return False

    # construct the enforcer
    if issuer is None:  # borrow the issuer from the token itself
        issuer = token["iss"]
    enforcer = Enforcer(issuer, audience=audience)

    # add validator for timeleft
    def _validate_timeleft(value):
        exp = float(value)
        return exp >= enforcer._now + timeleft

    enforcer.add_validator("exp", _validate_timeleft)

    # if scope wasn't given, borrow one from the token to pass validation
    if scope is None:
        scope = token["scope"].split(" ", 1)[:1]

    # iterate over given scopes and test all of them
    if isinstance(scope, str):
        scope = scope.split(" ")
    for scp in scope:
        # parse scope as scheme:path
        try:
            authz, path = scp.split(":", 1)
        except ValueError:
            authz = scp
            path = None

        # test
        try:
            res = enforcer.test(token, authz, path=path)
        except InvalidAuthorizationResource as exc:
            # bad scope in the token, is invalid
            res = False
            msg = f"{type(exc).__name__}: {exc}"
        else:
            msg = enforcer.last_failure
        if not res:
            if warn:
                warnings.warn(msg)
            return False

    return True


def target_audience(url, include_any=True):
    """Return the expected ``aud`` claim to authorize a request to ``url``.

    Parameters
    ----------
    url : `str`
        The URL that will be requested.

    include_any : `bool`, optional
        If `True`, include ``"ANY"`` in the return list of
        audiences, otherwise, don't.

    Returns
    -------
    audiences : `list` of `str`
        A `list` of audience values (`str`), either of length 1
        if ``include_any=False`, otherwise of length 2.

    Examples
    --------
    >>> default_audience(
    ...     "https://datafind.ligo.org:443/LDR/services/data/v1/gwf.json",
    ...     include_any=True,
    ... )
    ["https://datafind.ligo.org", "ANY"]
    >>> default_audience(
    ...     "segments.ligo.org",
    ...     include_any=False,
    ... )
    ["https://segments.ligo.org"]

    Hostnames given without a URL scheme are presumed to be HTTPS:

    >>> default_audience("datafind.ligo.org")
    ["https://datafind.ligo.org"]
    """
    if "//" not in url:  # always match a hostname, not a path
        url = f"//{url}"
    parsed = urlparse(url, scheme="https")
    aud = [f"{parsed.scheme}://{parsed.hostname}"]
    if include_any:
        aud.append("ANY")
    return aud


# -- I/O --------------------

def deserialize_token(raw, **kwargs):
    """Deserialize a token.

    Parameters
    ----------
    raw : `str`
        the raw serialised token content to deserialise

    kwargs
        all keyword arguments are passed on to
        :meth:`scitokens.SciToken.deserialize`

    Returns
    -------
    token : `scitokens.SciToken`
        the deserialised token

    See also
    --------
    scitokens.SciToken.deserialize
        for details of the deserialisation, and any valid keyword arguments
    """
    return SciToken.deserialize(raw.strip(), **kwargs)


def load_token_file(path, **kwargs):
    """Load a SciToken from a file path.

    Parameters
    ----------
    path : `str`
        the path to the scitokens file

    kwargs
        all keyword arguments are passed on to :func:`deserialize_token`

    Returns
    -------
    token : `scitokens.SciToken`
        the deserialised token

    Examples
    --------
    To load a token and validate a specific audience:

    >>> load_token('mytoken', audience="my.service.org")

    See also
    --------
    scitokens.SciToken.deserialize
        for details of the deserialisation, and any valid keyword arguments
    """
    with open(path, "r") as fobj:
        return deserialize_token(fobj.read(), **kwargs)


# -- discovery --------------

def find_token(
    audience,
    scope,
    issuer=None,
    timeleft=600,
    skip_errors=True,
    warn=False,
    **kwargs,
):
    """Find and load a `SciToken` for the given ``audience`` and ``scope``.

    Parameters
    ----------
    audience : `str`
        the required audience (``aud``).

    scope : `str`
        the required scope (``scope``).

    issuer : `str`
        the value of the `iss` claim to enforce.

    timeleft : `int`
        minimum required time left until expiry (in seconds)
        for a token to be considered 'valid'

    skip_errors : `bool`, optional
        skip over errors encoutered when attempting to deserialise
        discovered tokens; this may be useful to skip over invalid
        or expired tokens that exist, for example, which is why it
        is the default behaviour.

    warn : `bool`
        emit a warning when a token fails to deserialize, or fails
        validation.

    kwargs
        all keyword arguments are passed on to
        :meth:`scitokens.SciToken.deserialize`

    Returns
    -------
    token : `scitokens.SciToken`
        the first token that matches the requirements

    Raises
    ------
    ~igwn_auth_utils.IgwnAuthError
        if no valid token can be found

    See also
    --------
    scitokens.SciToken.deserialize
        for details of the deserialisation, and any valid keyword arguments
    """
    # preserve error from parsing tokens
    error = None

    # iterate over all of the tokens we can find for this audience
    for token in _find_tokens(audience=audience, **kwargs):
        # parsing a token yielded an exception, handle it here:
        if isinstance(token, Exception):
            error = error or token  # record (first) error for later
            if warn:  # emit a warning
                warnings.warn(f"{type(error).__name__}: {error}")
            if skip_errors:
                continue  # move on
            raise IgwnAuthError(str(error)) from error  # stop here and raise

        # if this token is valid, stop here and return it
        if is_valid_token(
            token,
            audience,
            scope,
            issuer=issuer,
            timeleft=timeleft,
            warn=warn,
        ):
            return token

    # if we didn't find any valid tokens:
    raise IgwnAuthError(
        "could not find a valid SciToken, "
        "please verify the audience and scope, "
        "or generate a new token and try again",
    ) from error


def _find_tokens(**deserialize_kwargs):
    """Yield all tokens that we can find

    This function will `yield` exceptions that are raised when
    attempting to parse a token that was actually found, so that
    they can be handled by the caller.
    """
    def _token_or_exception(func, *args, **kwargs):
        try:
            return func(*args, **kwargs)
        except TOKEN_ERROR as exc:
            return exc

    # read token directly from 'SCITOKEN{_FILE}' variable
    for envvar, loader in (
        ('SCITOKEN', deserialize_token),
        ('SCITOKEN_FILE', load_token_file),
    ):
        if envvar in os.environ:
            yield _token_or_exception(
                loader,
                os.environ[envvar],
                **deserialize_kwargs,
            )

    # try and find a token from HTCondor
    for tokenfile in _find_condor_creds_token_paths():
        yield _token_or_exception(
            load_token_file,
            tokenfile,
            **deserialize_kwargs,
        )

    try:
        yield _token_or_exception(SciToken.discover, **deserialize_kwargs)
    except OSError:  # no token
        pass  # try something else
    except AttributeError as exc:
        # windows doesn't have geteuid, that's ok, otherwise panic
        if not WINDOWS or "geteuid" not in str(exc):
            raise


def _find_condor_creds_token_paths():
    """Find all token files in the condor creds directory
    """
    try:
        _condor_creds_dir = Path(os.environ["_CONDOR_CREDS"])
    except KeyError:
        return
    try:
        for f in _condor_creds_dir.iterdir():
            if f.suffix == ".use":
                yield f
    except FileNotFoundError:   # creds dir doesn't exist
        return


# -- HTTP request helper ----

def token_authorization_header(token, scheme="Bearer"):
    """Format an in-memory token for use in an HTTP Authorization Header.

    Parameters
    ----------
    token : `scitokens.SciToken`
        the token to format

    scheme : `str` optional
        the Authorization scheme to use

    Returns
    -------
    header_str : `str`
        formatted content for an `Authorization` header

    Notes
    -----
    See `RFC-6750 <https://datatracker.ietf.org/doc/html/rfc6750>`__
    for details on the ``Bearer`` Authorization token standard.
    """
    return "{} {}".format(
        scheme,
        token._serialized_token or token.serialize().decode("utf-8"),
    )



# -*- coding: utf-8 -*-
# Copyright 2021-2022 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Tests for :mod:`igwn_auth_utils.requests`.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Leo Singer <leo.singer@ligo.org>"

import os
import stat
from netrc import NetrcParseError
from unittest import mock
from urllib.parse import urlencode

import pytest

from requests import (
    __version__ as requests_version,
    RequestException,
)

from .. import requests as igwn_requests
from ..error import IgwnAuthError
from ..scitokens import target_audience
from .test_scitokens import rtoken  # noqa: F401

SKIP_REQUESTS_NETRC = pytest.mark.skipif(
    requests_version < "2.25.0",
    reason=f"requests {requests_version} doesn't respect NETRC env",
)


# -- utilities ------------------------

def _empty(*args, **kwargs):
    return []


def _igwnerror(*args, **kwargs):
    raise IgwnAuthError("error")


def mock_no_scitoken():
    return mock.patch(
        "igwn_auth_utils.scitokens._find_tokens",
        _empty,
    )


def mock_no_x509():
    return mock.patch(
        "igwn_auth_utils.requests.find_x509_credentials",
        _igwnerror,
    )


@pytest.fixture
def netrc(tmp_path):
    netrc = tmp_path / "netrc"
    netrc.write_text(
        "machine example.org login albert.einstein password super-secret",
    )
    netrc.chmod(stat.S_IRWXU)
    return netrc


def has_auth(session):
    return bool(
        session.auth
        or session.cert
        or "Authorization" in session.headers
    )


class MockRequest(mock.MagicMock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.headers = {}


# -- get_netrc_auth -------------------

@mock.patch.dict(os.environ)
@SKIP_REQUESTS_NETRC
def test_get_netrc_auth(netrc):
    os.environ["NETRC"] = str(netrc)
    assert igwn_requests.get_netrc_auth("https://example.org/path") == (
        "albert.einstein",
        "super-secret",
    )


@mock.patch.dict(os.environ)
@SKIP_REQUESTS_NETRC
def test_get_netrc_auth_nomatch(netrc):
    os.environ["NETRC"] = str(netrc)
    assert igwn_requests.get_netrc_auth("https://bad.org/path") is None


@mock.patch.dict(os.environ)
@SKIP_REQUESTS_NETRC
def test_get_netrc_auth_notfound(tmp_path):
    netrc_file = tmp_path / "netrc"
    os.environ["NETRC"] = str(netrc_file)
    # file not found
    assert igwn_requests.get_netrc_auth(None) is None


@mock.patch.dict(os.environ)
@pytest.mark.skipif(
    os.name == "nt",
    reason="safe_netrc doesn't do anything on Windows",
)
@SKIP_REQUESTS_NETRC
def test_get_netrc_auth_permissions(netrc):
    os.environ["NETRC"] = str(netrc)
    netrc.chmod(0o444)
    with pytest.raises(NetrcParseError):
        igwn_requests.get_netrc_auth(None, raise_errors=True)
    assert igwn_requests.get_netrc_auth(None, raise_errors=False) is None


# -- HTTPSciTokenAuth -----------------

class TestHTTPSciTokenAuth:
    Auth = igwn_requests.HTTPSciTokenAuth

    def test_init(self):
        auth = self.Auth()
        assert auth.token is None
        assert auth.audience is None
        assert auth.scope is None
        assert auth.issuer is None

    def test_eq(self):
        a = self.Auth(token=None, audience="ANY")
        b = self.Auth(token=None, audience="ANY")
        assert a == b

    def test_neq(self):
        a = self.Auth(token=None, audience="ANY")
        b = self.Auth(token=None, audience="https://example.com")
        assert a != b

    @mock.patch("igwn_auth_utils.requests.find_scitoken")
    def test_token_header_empty(self, find_token):  # noqa: F811
        """Test that the auth class handles no tokens properly.
        """
        find_token.return_value = None
        req = MockRequest()
        auth = self.Auth()
        assert auth(req).headers.get("Authorization") is None

    @mock.patch("igwn_auth_utils.requests.find_scitoken")
    def test_token_header(self, find_token, rtoken):  # noqa: F811
        """Test that the auth class finds the token and serialises
        it into a header properly.
        """
        find_token.return_value = rtoken

        auth = self.Auth()
        req = MockRequest()
        assert auth(req).headers["Authorization"] == (
            igwn_requests.scitoken_authorization_header(rtoken)
        )


# -- Session --------------------------

class TestSession:
    Session = igwn_requests.Session

    # -- SessionErrorMixin

    def test_raise_for_status_hook(self, requests_mock):
        # define a request that returns 404 (not found)
        requests_mock.get(
            "https://test.org",
            status_code=404,
            reason="not found",
        )

        # with the kwarg a RequestException is raised
        with pytest.raises(
            RequestException,
            match=r"404 Client Error: not found for url: https://test.org/",
        ):
            igwn_requests.get("https://test.org")

    # -- session auth

    def test_noauth_args(self):
        """Test that `Session(force_noauth=True, fail_if_noauth=True)`
        is invalid.
        """
        with pytest.raises(ValueError):
            self.Session(force_noauth=True, fail_if_noauth=True)

    def test_fail_if_noauth(self):
        """Test that `Session(fail_if_noauth=True)` raises an error
        """
        with pytest.raises(IgwnAuthError):
            self.Session(
                token=False,
                cert=False,
                url=None,
                fail_if_noauth=True,
            )

    def test_force_noauth(self):
        """Test that `Session(force_noauth=True)` overrides auth kwargs
        """
        sess = self.Session(cert="cert.pem", force_noauth=True)
        assert sess.cert is False
        assert sess.auth is None

    @mock_no_scitoken()
    @mock_no_x509()
    def test_defaults(self):
        """Test that the `Session()` defaults work in a noauth environment
        """
        sess = self.Session()
        assert sess.cert is None
        assert isinstance(sess.auth, igwn_requests.HTTPSciTokenAuth)
        assert sess.auth.token is None

    # -- tokens

    def test_token_explicit(self, rtoken):  # noqa: F811
        """Test that tokens are handled properly
        """
        sess = self.Session(token=rtoken)
        assert sess.auth.token is rtoken
        # mock the request to get the header that would be used
        req = MockRequest()
        sess.auth(req)
        assert req.headers["Authorization"] == (
            igwn_requests.scitoken_authorization_header(rtoken)
        )

    def test_token_serialized(self, rtoken):  # noqa: F811
        """Test that serialized tokens are handled properly
        """
        serialized = rtoken.serialize()
        sess = self.Session(token=serialized)
        req = MockRequest()
        sess.auth(req)
        assert req.headers["Authorization"] == f"Bearer {serialized}"
        # will not deserialise a token for storage
        assert sess.token is None

    @mock.patch("igwn_auth_utils.requests.find_scitoken")
    def test_token_discovery(self, find_token, rtoken):  # noqa: F811
        find_token.return_value = rtoken
        sess = self.Session()
        sess.auth(sess)
        assert sess.headers["Authorization"] == (
            igwn_requests.scitoken_authorization_header(rtoken)
        )

    @mock.patch(
        "igwn_auth_utils.requests.find_scitoken",
        side_effect=IgwnAuthError,
    )
    def test_token_required_failure(self, _):
        with pytest.raises(IgwnAuthError), self.Session(token=True) as sess:
            sess.get("https://example.com")

    @pytest.mark.parametrize(("url", "aud"), (
        ("https://secret.example.com:8008", ["https://secret.example.com"]),
        (None, None)
    ))
    @mock.patch("igwn_auth_utils.requests.find_scitoken")
    def test_token_audience_default(self, find_scitoken, url, aud):
        """Check that the default `token_audience` is set correctly.
        """
        sess = self.Session(
            url=url,
            token=True,
            cert=False,
            auth=None,
        )
        assert sess.auth.audience == aud

    # -- X.509

    def test_cert_explicit(self):
        """Test that cert credentials are stored properly
        """
        sess = self.Session(token=False, cert="cert.pem")
        assert sess.cert == "cert.pem"
        assert sess.auth is None

    @mock.patch(
        "igwn_auth_utils.requests.find_x509_credentials",
        return_value="test.pem",
    )
    def test_cert_discovery(self, _):
        """Test that automatic certificate discovery works
        """
        assert self.Session(token=False).cert == "test.pem"

    @pytest.mark.skipif(
        not igwn_requests.REQUESTS_CERT_HTTP_BUG,
        reason=f"bug not present on requests {requests_version}",
    )
    def test_cert_requests_214(self):
        """Test that cert is disabled on requests 2.14 by default.
        """
        s = self.Session(url="http://example.com")
        assert s.cert is False

    @mock.patch(
        "igwn_auth_utils.requests.find_x509_credentials",
        side_effect=IgwnAuthError,
    )
    def test_cert_required_failure(self, _):
        with pytest.raises(IgwnAuthError):
            self.Session(token=False, cert=True)

    # -- basic auth

    @mock.patch.dict(os.environ)
    @pytest.mark.parametrize(("url", "auth"), (
        ("https://example.org", ("albert.einstein", "super-secret")),
        ("https://bad.org", None),
    ))
    @SKIP_REQUESTS_NETRC
    def test_basic_auth(self, netrc, url, auth):
        os.environ["NETRC"] = str(netrc)
        sess = self.Session(cert=False, token=False, url=url)
        assert sess.auth == auth

    # -- all

    @mock.patch("igwn_auth_utils.requests.find_scitoken")
    @mock.patch(
        "igwn_auth_utils.requests.find_x509_credentials",
        return_value=None,
    )
    @pytest.mark.parametrize(("cert", "token", "auth"), (
        (None, None, None),  # none
        ("A", False, ("C", "D")),  # no token
        ("A", True, None),  # no basic auth
        (None, True, None),  # no cert or basic auth
        (False, False, ("C", "D")),  # no cert or token
    ))
    def test_multi_auth(
        self,
        find_x509,
        find_token,
        rtoken,  # noqa: F811
        cert,
        token,
        auth,
    ):
        """Check that Session._init_auth records all auth options

        In case a remote host accepts X.509 but not tokens, but the user
        has a valid ANY token (for example).
        """
        find_token.return_value = rtoken
        sess = self.Session(cert=cert, token=token, auth=auth)
        assert sess.cert == cert
        if token:
            sess.auth(sess)
            sess.headers["Authorization"].startswith("Bearer")
        else:
            assert sess.auth == auth

    # -- request auth
    # test that Session auth and Request auth play nicely together

    @mock.patch(
        "igwn_auth_utils.x509.is_valid_certificate",
        return_value=False,
    )
    def test_request_x509(self, is_valid, requests_mock, tmp_path):
        """Test that a request does its own search for an X.509 cert.
        """
        x509 = tmp_path / "x509"
        requests_mock.get("https://example.com")
        with mock.patch.dict(os.environ):
            os.environ.pop("X509_USER_PROXY", None)
            with self.Session(cert=None) as sess:
                # check that the Session doesn't have a cert
                assert sess.cert is None

                # but that the request uses one because it calls
                # find_x509_credentials again
                x509.touch()
                os.environ["X509_USER_PROXY"] = str(x509)
                resp = sess.get("https://example.com", cert=None)
                assert resp.request.cert == str(x509)

    @mock.patch("igwn_auth_utils.requests.find_scitoken", return_value=None)
    @pytest.mark.parametrize(
        ("session_aud", "session_scope", "request_aud", "request_scope"),
        [
            (None, None, "aud", "scope"),
            ("aud", "scope", None, "newscope"),
            ("aud", "scope", "newaud", None),
            (None, None, None, None),
        ],
    )
    def test_request_token_auth(
        self,
        find_scitoken,
        requests_mock,
        session_aud,
        session_scope,
        request_aud,
        request_scope,
    ):
        """Test that a request correctly merges token claim settings.
        """
        requests_mock.get("https://example.com/api")
        with self.Session(
            cert=False,
            token_audience=session_aud,
            token_scope=session_scope,
        ) as sess:  # use a token
            # check that the session auth handler recorded what we gave it
            assert sess.auth.audience == session_aud
            assert sess.auth.scope == session_scope

            audience = (
                request_aud
                or session_aud
                or target_audience(
                    "https://example.com/api",
                    include_any=False,
                )
            )

            # but that the request auth uses any new settings we give it
            sess.get(
                "https://example.com/api",
                token_audience=request_aud,
                token_scope=request_scope,
            )
            find_scitoken.assert_called_once_with(
                audience,
                request_scope or session_scope,
                issuer=None,
            )

    @mock.patch("igwn_auth_utils.requests.find_scitoken", return_value=None)
    def test_request_fail_if_noauth(self, find_scitoken):
        """Test that `Session.get(fail_if_noauth=True)` raises an error.
        """
        with self.Session(fail_if_noauth=False) as sess, \
             pytest.raises(IgwnAuthError, match="no valid authorisation"):
            sess.get(
                "https://example.com",
                cert=False,
                fail_if_noauth=True,
            )

    @mock.patch("igwn_auth_utils.requests.HTTPSciTokenAuth.__call__")
    def test_request_token_false(self, token_auth_call, requests_mock):
        """Test that token=False on the request is respected.

        Regression: https://git.ligo.org/computing/igwn-auth-utils/-/issues/14
        """
        requests_mock.get("https://example.com/api")
        with self.Session(cert=False, raise_for_status=False) as sess:
            assert isinstance(sess.auth, igwn_requests.HTTPSciTokenAuth)
            sess.get("https://example.com/api", token=False)
            # assert that token=False is propagated properly
            # so that HTTPSciTokenAuth is never actually invoked
            token_auth_call.assert_not_called()

            # check that if we want a token the next time, it gets used
            sess.get("https://example.com/api")
            token_auth_call.assert_called_once()


# -- standalone requests --------------

def test_get(requests_mock):
    """Test that `igwn_auth_utils.requests.get` can perform a simple request
    """
    requests_mock.get(
        "https://test.org",
        text="TEST",
    )
    assert igwn_requests.get("https://test.org").text == "TEST"


@mock.patch("igwn_auth_utils.requests.HTTPSciTokenAuth.__call__")
def test_get_token_false(token_auth_call, requests_mock):
    """Test that `igwn_auth_utils.requests.get` respects token=False.
    """
    requests_mock.get(
        "https://test.org",
        text="TEST",
    )
    igwn_requests.get("https://test.org", token=False)
    # assert that HTTPSciTokenAuth is never actually invoked
    token_auth_call.assert_not_called()


@mock.patch("igwn_auth_utils.requests.Session")
def test_get_session(mock_session):
    """Test that ``session`` for `igwn_auth_utils.requests.get` works
    """
    session = mock.MagicMock()
    assert igwn_requests.get("https://test.org", session=session)
    session.request.assert_called_once_with("get", "https://test.org")
    mock_session.assert_not_called()


@mock.patch("igwn_auth_utils.requests.find_scitoken")
@mock.patch("igwn_auth_utils.requests.find_x509_credentials")
def test_get_force_noauth(find_x509, find_scitoken, requests_mock):
    """Test that `igwn_auth_utils.requests.get` passes `force_noauth` properly.

    Regression: <https://git.ligo.org/computing/igwn-auth-utils/-/issues/12>
    """
    requests_mock.get("https://test.org")
    igwn_requests.get("https://test.org", force_noauth=True)
    find_x509.assert_not_called()
    find_scitoken.assert_not_called()


def test_post(requests_mock):
    """Test that `igwn_auth_utils.requests.post` can perform a simple request.
    """
    data = {"a": 1, "b": 2}
    requests_mock.post(
        "https://example.com",
        text="THANKS",
    )
    # check that the correct response got passed through
    assert igwn_requests.post(
        "https://example.com",
        data=data,
    ).text == "THANKS"
    # check that the data was encoded into the request properly
    req = requests_mock.request_history[0]
    assert req.body == urlencode(data)



# -*- coding: utf-8 -*-
# Copyright 2021-2023 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Python Requests interface with IGWN authentication.

This is heavily inspired by Leo Singer's excellent
:mod:`requests_gracedb` package.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Leo Singer <leo.singer@ligo.org>"

import sys
from functools import wraps
from unittest import mock

import requests
from requests.auth import AuthBase as _AuthBase
from requests import utils as requests_utils

from scitokens import SciToken

from .error import IgwnAuthError
from .scitokens import (
    find_token as find_scitoken,
    target_audience as scitoken_audience,
    token_authorization_header as scitoken_authorization_header,
)
from .x509 import (
    find_credentials as find_x509_credentials,
)

# handle https://github.com/psf/requests/issues/4025
# NOTE: 2.14.1 is the base version on EPEL7
REQUESTS_CERT_HTTP_BUG = '2.14.1' <= requests.__version__ < '2.15.0'


# -- Auth utilities -------------------

def _find_cred(func, *args, error=True, **kwargs):
    """Find a credential and maybe ignore an `~igwn_auth_utils.IgwnAuthError`

    This is an internal utility for the `SessionAuthMixin._init_auth`
    method which shouldn't necessary fail if it doesn't
    find a credential of any one type, but should just move on to the
    next option.
    """
    try:
        return func(*args, **kwargs)
    except IgwnAuthError:
        if error:
            raise
        return


@wraps(requests_utils.get_netrc_auth)
def get_netrc_auth(url, raise_errors=False):
    """Wrapper around `requests.utils.get_netrc_auth` to use `safe_netrc`.
    """
    import safe_netrc
    with mock.patch.dict(sys.modules, {"netrc": safe_netrc}):
        return requests_utils.get_netrc_auth(url, raise_errors=raise_errors)


class HTTPSciTokenAuth(_AuthBase):
    """Auth handler for SciTokens.
    """
    def __init__(
        self,
        token=None,
        audience=None,
        scope=None,
        issuer=None,
    ):
        self.token = token
        self.audience = audience
        self.scope = scope
        self.issuer = issuer

    def __eq__(self, other):
        return all([
            self.token == getattr(other, "token", None),
            self.audience == getattr(other, "audience", None),
            self.scope == getattr(other, "scope", None),
            self.issuer == getattr(other, "issuer", None),
        ])

    def __ne__(self, other):
        return not self == other

    @staticmethod
    def _auth_header_str(token):
        """Serialise a `scitokens.SciToken` and format an Authorization header.

        Parameters
        ----------
        token : `scitokens.SciToken`, `str`, `bytes`
            the token to serialize, or an already serialized representation
        """
        if isinstance(token, (str, bytes)):  # already serialised
            return f"Bearer {token}"
        return scitoken_authorization_header(token)

    def find_token(
        self,
        url=None,
        error=True,
    ):
        """Find a bearer token for authorization.

        Parameters
        ----------
        url : `str`
            The URL that will be queried.

        error : `bool`
            If `True`, `raise` exceptions, otherwise return `None`.
        """
        audience = self.audience
        if audience is None and url is not None:
            audience = scitoken_audience(url, include_any=False)
        return _find_cred(
            find_scitoken,
            audience,
            self.scope,
            issuer=self.issuer,
            error=error,
        )

    def __call__(self, r):
        """Augment the `Request` ``r`` with an ``Authorization`` header.
        """
        token = self.token
        if token in (None, True):
            token = self.find_token(
                url=getattr(r, "url", None),  # allow r as Session
                error=bool(token),
            )

        # if we ended up with a header, store it in the request.
        if token:
            r.headers["Authorization"] = self._auth_header_str(token)

        return r


def _prepare_auth(
    url=None,
    auth=None,
    cert=None,
    token=None,
    token_audience=None,
    token_scope=None,
    token_issuer=None,
    force_noauth=False,
    fail_if_noauth=False,
    session=None,
):
    """Prepare authorisation for a session or request.
    """
    # workaround https://github.com/psf/requests/issues/4025
    if (
        cert is None
        and str(url).startswith("http://")
        and REQUESTS_CERT_HTTP_BUG
    ):
        cert = False

    # merge settings from the session
    if session:
        if cert is None:
            cert = session.cert
        if token is None and session.auth is None:
            token = False
        if isinstance(session.auth, HTTPSciTokenAuth):
            if token is None:
                token = session.auth.token
            if token_audience is None:
                token_audience = session.auth.audience
            if token_scope is None:
                token_scope = session.auth.scope
            if token_issuer is None:
                token_issuer = session.auth.issuer

    # handle options
    if force_noauth and fail_if_noauth:
        raise ValueError(
            "cannot select both force_noauth and fail_if_noauth",
        )
    if force_noauth:
        return None, False

    # cert auth (always attach if we can)
    if cert in (None, True):  # not disabled and not given explicitly
        cert = _find_cred(find_x509_credentials, error=cert is True)

    # use existing auth object
    if auth is not None:
        pass

    # -- bearer token (scitoken)

    elif token is not False:
        # get the default audience from the URL
        if token_audience is None and url is not None:
            token_audience = scitoken_audience(url, include_any=False)
        auth = HTTPSciTokenAuth(
            token=token,
            audience=token_audience,
            scope=token_scope,
            issuer=token_issuer,
        )

    # -- basic auth (netrc)

    elif (
        # we know where the requests are heading
        url is not None
        # and basic auth not disabled
        and auth in (None, True)
    ):
        auth = get_netrc_auth(url, raise_errors=False)

    # -- handle session/request auth combination

    if (
        # if the parent Session is configured for token auth
        session is not None and isinstance(session.auth, HTTPSciTokenAuth)
        # but _this request_ came with token=False
        and token is False
        # and didn't get configured for any other auth
        and auth is None
    ):
        # we need to forcibly disable using the session's token auth
        auth = False

    # -- handle fail_if_noauth

    # if no auth was found, and we need it, fail here
    if fail_if_noauth and cert in (None, False) and (
        not auth
        or (
            isinstance(auth, HTTPSciTokenAuth)
            and not auth.find_token(url=url, error=False)
        )
    ):
        raise IgwnAuthError("no valid authorisation credentials found")

    return auth, cert


# -- Session handling -----------------

_auth_session_parameters = """
    Discovery/configuration of authorisation/authentication methods
    is attempted in the following order:

    1.  if ``force_noauth=True`` is given, no auth is configured;

    2.  for SciTokens:

        1.  if a bearer token is provided via the ``token`` keyword argument,
            then use that, or

        2.  look for a bearer token by passing the ``token_audience``
            and ``token_scope`` keyword parameters to
            :func:`igwn_auth_utils.find_scitokens`;

    3.  for X.509 credentials:

        1.  if an X.509 credential path is provided via the ``cert`` keyword
            argument, then use that, or

        2.  look for an X.509 credential using
            :func:`igwn_auth_utils.find_x509_credential`

    4.  for basic auth (username/password):

        1.  if ``auth`` keyword is provided, then use that, or

        2.  read the netrc file located at :file:`~/.netrc`, or at the path
            stored in the :envvar:`$NETRC` environment variable, and look
            for a username and password matching the hostname given in the
            ``url`` keyword argument;

    5.  if none of the above yield a credential, and ``fail_if_noauth=True``
        was provided, raise a `ValueError`.

    Steps 2 and 3 are all tried independently, with all valid credentials
    (one per type) configured for the session.
    Only when SciTokens are disabled (``token=False``), will step 4 will be
    tried to configure basic username/password auth.
    It is up to the request receiver to handle the multiple credential
    types and prioritise between them.

    Parameters
    ----------
    token : `scitokens.SciToken`, `str`, `bool`, optional
        Bearer token (scitoken) input, one of

        - a bearer token (`scitokens.SciToken`),
        - a serialised token (`str`, `bytes`),
        - `False`: disable using tokens completely
        - `True`: discover a valid token via
          :func:`igwn_auth_utils.find_scitoken` and
          error if something goes wrong
        - `None`: try and discover a valid token, but
          try something else if that fails

    token_audience : `str`, list` of `str`
        The value(s) of the audience (``aud``) claim to pass to
        :func:`igwn_auth_utils.find_scitoken` when discovering
        available tokens.

    token_scope : `str`
        The value(s) of the ``scope`` to pass to
        :func:`igwn_auth_utils.find_scitoken` when discovering
        available tokens.

    token_issuer : `str`
        The value of the issuer (``iss``) claim to pass to
        :func:`igwn_auth_utils.find_scitoken` when discovering
        available tokens.

    cert : `str`, `tuple`, `bool`, optional
        X.509 credential input, one of

        - path to a PEM-format certificate file,
        - a ``(cert, key)`` `tuple`,
        - `False`: disable using X.509 completely
        - `True`: discover a valid cert via
          :func:`igwn_auth_utils.find_x509_credentials` and
          error if something goes wrong
        - `None`: try and discover a valid cert, but
          try something else if that fails

    auth :  `tuple`, `object`, optional
        ``(username, password)`` `tuple` or other authentication/authorization
        object to attach to a `~requests.Request`.
        By default a new :class:`HTTPSciTokenAuth` handler will be attached
        to configure ``Authorization`` headers for each request.

    url : `str`, optional
        the URL/host that will be queried within this session; this is used
        to set the default ``token_audience`` and to access credentials
        via :mod:`safe_netrc`.

    force_noauth : `bool`, optional
        Disable the use of any authorisation credentials (mainly for testing).

    fail_if_noauth : `bool`, optional
        Raise a `~igwn_auth_utils.IgwnAuthError` if no authorisation
        credentials are presented or discovered.

    raise_for_status : `bool`, optional
        If `True` (default), automatically call
        :meth:`~requests.Response.raise_for_status` after receiving
        any response.

    Raises
    ------
    ~igwn_auth_utils.IgwnAuthError
        If ``cert=True`` or ``token=True`` is given and the relevant
        credential was not actually discovered, or
        if ``fail_if_noauth=True`` is given and no authorisation
        token/credentials of any valid type are presented or discovered.

    See also
    --------
    requests.Session
        for details of the standard options

    igwn_auth_utils.find_scitoken
        for details of the SciToken discovery

    igwn_auth_utils.find_x509_credentials
        for details of the X.509 credential discovery
    """.strip()


def _hook_raise_for_status(response, *args, **kwargs):
    """Response hook to raise exception for any HTTP error (status >= 400)

    Reproduced (with permission) from :mod:`requests_gracedb.errors`,
    authored by Leo Singer.
    """
    return response.raise_for_status()


class SessionErrorMixin:
    """A mixin for :class:`requests.Session` to raise exceptions for HTTP
    errors.

    Reproduced (with permission) from :mod:`requests_gracedb.errors`,
    authored by Leo Singer.
    """
    def __init__(self, *args, **kwargs):
        raise_for_status = kwargs.pop("raise_for_status", True)
        super().__init__(*args, **kwargs)
        if raise_for_status:
            self.hooks.setdefault("response", []).append(
                _hook_raise_for_status,
            )


class SessionAuthMixin:
    """Mixin for :class:`requests.Session` to add support for IGWN auth.

    By default this mixin will automatically attempt to discover/configure
    a bearer token (scitoken) or an X.509 credential, with options to
    require/disable either of those, or all authentication entirely.

    {parameters}
    """
    def __init__(
        self,
        token=None,
        token_audience=None,
        token_scope=None,
        token_issuer=None,
        cert=None,
        auth=None,
        url=None,
        force_noauth=False,
        fail_if_noauth=False,
        **kwargs,
    ):
        # initialise session
        super().__init__(**kwargs)

        # initialise auth handler and cert
        self._init_auth(
            url=url,
            auth=auth,
            cert=cert,
            token=token,
            token_audience=token_audience,
            token_scope=token_scope,
            token_issuer=token_issuer,
            force_noauth=force_noauth,
            fail_if_noauth=fail_if_noauth,
        )

    def _init_auth(self, url=None, token=None, **kwargs):
        """Initialise the auth handler for this `Session`.
        """
        # find creds if we can
        self.auth, self.cert = _prepare_auth(url=url, token=token, **kwargs)

        # if we were given a token, use it now
        if isinstance(token, (SciToken, str, bytes)):
            self.auth(self)

    @property
    def token(self):
        """The token object that will be used in authorised requests.

        If the :attr:`Session.auth` property isn't an instance of
        `HTTPSciTokenAuth` with a token attached, this returns `None`.
        """
        token = getattr(self.auth, "token", None)
        if isinstance(token, SciToken):
            return token


class Session(
    SessionAuthMixin,
    SessionErrorMixin,
    requests.Session,
):
    """`requests.Session` class with default IGWN authorization handling.

    {parameters}

    Examples
    --------
    To use the default authorisation discovery:

    >>> from igwn_auth_utils.requests import Session
    >>> with Session() as sess:
    ...     sess.get("https://science.example.com/api/important/data")

    To explicitly pass a specific :class:`~scitokens.SciToken` as the token:

    >>> with Session(token=mytoken) as sess:
    ...     sess.get("https://science.example.com/api/important/data")

    To explicitly *require* that a token is discovered, and *disable*
    any X.509 discovery:

    >>> with Session(token=True, x509=False) as sess:
    ...     sess.get("https://science.example.com/api/important/data")

    To use default authorisation discovery, but fail if no credentials
    are discovered:

    >>> with Session(fail_if_noauth=True) as sess:
    ...     sess.get("https://science.example.com/api/important/data")

    To disable all authorisation discovery:

    >>> with Session(force_noauth=True) as sess:
    ...     sess.get("https://science.example.com/api/important/data")
    """
    __attrs__ = requests.Session.__attrs__ = [
        "token",
    ]

    @wraps(requests.Session.request)
    def request(
        self,
        method,
        url,
        *args,
        token=None,
        token_audience=None,
        token_scope=None,
        token_issuer=None,
        cert=None,
        auth=None,
        force_noauth=False,
        fail_if_noauth=False,
        **kwargs,
    ):
        # handle request-specific auth
        auth, cert = _prepare_auth(
            url=url,
            auth=auth,
            cert=cert,
            token=token,
            token_audience=token_audience,
            token_scope=token_scope,
            token_issuer=token_issuer,
            force_noauth=force_noauth,
            fail_if_noauth=fail_if_noauth,
            session=self,
        )

        # continue with request
        return super().request(
            method,
            url,
            *args,
            auth=auth,
            cert=cert,
            **kwargs,
        )


# update the docstrings to include the same parameter info
for _obj in (Session, SessionAuthMixin):
    _obj.__doc__ = _obj.__doc__.format(parameters=_auth_session_parameters)


# -- standalone request handling ------

def request(method, url, *args, session=None, **kwargs):
    """Send a request of the specific method to the specified URL.

    Parameters
    ----------
    method : `str`
        The method to use.

    url : `str`,
        The URL to request.

    session : `requests.Session`, optional
        The connection session to use, if not given one will be
        created on-the-fly.

    args, kwargs
        All other keyword arguments are passed directly to
        `requests.Session.request`

    Returns
    -------
    resp : `requests.Response`
        the response object

    See also
    --------
    igwn_auth_utils.requests.Session.request
        for information on how the request is performed
    """
    # user's session
    if session:
        return session.request(method, url, *args, **kwargs)

    # give the Session constructor everything for auth as well
    session_kw = {
        k: kwargs[k] for k in (
            "auth",
            "cert",
            "fail_if_noauth",
            "force_noauth",
            "token",
            "token_audience",
            "token_issuer",
            "token_scope",
        ) if k in kwargs
    }

    # new session
    with Session(url=url, **session_kw) as session:
        return session.request(method, url, *args, **kwargs)


_request_wrapper_doc = """
    Send an HTTP {METHOD} request to the specified URL with IGWN Auth attached.

    Parameters
    ----------
    url : `str`
        The URL to request.

    session : `requests.Session`, optional
        The connection session to use, if not given one will be
        created on-the-fly.

    args, kwargs
        All other keyword arguments are passed directly to
        :meth:`requests.Session.{method}`

    Returns
    -------
    resp : `requests.Response`
        the response object

    See also
    --------
    requests.Session.{method}
        for information on how the request is performed
""".strip()


def _request_wrapper_factory(method):
    """Factor function to wrap a :mod:`requests` HTTP method to use
    our request function.
    """
    def _request_wrapper(url, *args, session=None, **kwargs):
        return request(method, url, *args, session=session, **kwargs)

    _request_wrapper.__doc__ = _request_wrapper_doc.format(
        method=method,
        METHOD=method.upper(),
    )
    return _request_wrapper


# request methods
delete = _request_wrapper_factory("delete")
get = _request_wrapper_factory("get")
head = _request_wrapper_factory("head")
patch = _request_wrapper_factory("patch")
post = _request_wrapper_factory("post")
put = _request_wrapper_factory("put")
